#import "AppDelegate.h"

#import "BookmarkEntity.h"
#import "FolderEntity.h"

#define kBookmarksSchemaPath @"/System/Library/SyncServices/Schemas/Bookmarks.syncschema"
#define kRecordEntityName @"com.apple.syncservices.RecordEntityName"

@interface AppDelegate()

@property (retain) NSMutableDictionary *recordLookup;
@property (retain) NSProgressIndicator* progressIndicator;
@property (assign) ISyncSessionDriverMode preferredSyncMode;

-(NSProgressIndicator*)progressIndicator;

@end

@implementation AppDelegate

@synthesize bookmarkObjects = _bookmarkObjects;
@synthesize progressIndicator = _progressIndicator;
@synthesize preferredSyncMode = _preferredSyncMode;
@synthesize recordLookup = _recordLookup;

-(id)init
{
	if (![super init]) return nil;
  
  [self setRecordLookup:[NSMutableDictionary dictionary]];
	
	syncDriver = [[ISyncSessionDriver sessionDriverWithDataSource:self] retain];
  [syncDriver setDelegate:self];
	
	[self setPreferredSyncMode:ISyncSessionDriverModeFast];
  
	SEL syncSEL = @selector(client:willSyncEntityNames:);
  [[syncDriver client] setSyncAlertHandler:self selector:syncSEL];
  
  [syncDriver sync];
	
	
	return self;
}

- (void)client:(ISyncClient*)client willSyncEntityNames:(NSArray*)entityNames
{
  [syncDriver sync];
}

- (id)objectForRecordIdentifier:(NSString*)identifier;
{
  return [[self recordLookup] objectForKey:identifier];
}

- (void)applicationWillTerminate:(NSNotification *)aNotification
{
	//Debug purposes since we are not saving the data
	[[ISyncManager sharedManager] unregisterClient:[syncDriver client]];
}

- (IBAction)requestSync:(id)sender
{
	[syncDriver sync];
}

#pragma mark -
#pragma mark Sync Delegate methods

- (void)sessionDriverDidFinishSession:(ISyncSessionDriver *)sender
{
  NSMutableArray *topLevelObjects = [NSMutableArray array];
	
  for (NSString* key in [[self recordLookup] allKeys]) {
    id object = [[self recordLookup] objectForKey:key];
    if ([object parent]) continue;
    [topLevelObjects addObject:object];
  }
  [self setBookmarkObjects:topLevelObjects];
}

- (void)sessionDriverDidCancelSession:(ISyncSessionDriver*)sender
{
  NSLog(@"%@:%s entered", [self class], _cmd);
}

#pragma mark -
#pragma mark ISyncSessionDriver methods

-(NSString*)clientIdentifier
{
	return [[NSBundle mainBundle] bundleIdentifier];
}

-(NSArray*)schemaBundleURLs 
{
	NSURL *schemaURL = [NSURL fileURLWithPath:kBookmarksSchemaPath];
	return [NSArray arrayWithObject:schemaURL];
}

-(NSArray*)entityNamesToSync
{
	return [NSArray arrayWithObjects:[FolderEntity entityName], 
          [BookmarkEntity entityName], nil];
}

-(ISyncSessionDriverMode)preferredSyncModeForEntityName:(NSString*)entity
{
	return [self preferredSyncMode];
}

-(BOOL)deleteAllRecordsForEntityName:(NSString*)entityName 
                               error:(NSError**)outError
{
  for (NSString *key in [[self recordLookup] allValues]) {
    id object = [[self recordLookup] objectForKey:key];
    if (![[object entityName] isEqualToString:entityName]) {
      continue;
    }
    [[self recordLookup] removeObjectForKey:key];
  }
	return YES;
}

-(ISyncSessionDriverChangeResult)applyChange:(ISyncChange*)change 
																forEntityName:(NSString*)entityName 
										 remappedRecordIdentifier:(NSString**)outRecordIdentifier 
															formattedRecord:(NSDictionary**)outRecord 
																				error:(NSError**)err
{
  NSDictionary *errDict = nil;
  id record = nil;
  switch ([change type]) {
    case ISyncChangeTypeDelete:
      [[self recordLookup] removeObjectForKey:[change recordIdentifier]];
      return ISyncSessionDriverChangeAccepted;
    case ISyncChangeTypeAdd:
      if ([entityName isEqualToString:[FolderEntity entityName]]) {
        record = [[FolderEntity alloc] init];
      } else if ([entityName isEqualToString:[BookmarkEntity entityName]]) {
        record = [[BookmarkEntity alloc] init];
      } else {
        errDict = [NSDictionary dictionaryWithObject:@"Unknown type"
                                              forKey:NSLocalizedDescriptionKey];
        *err = [NSError errorWithDomain:@"CIMGF" code:8002 userInfo:errDict];
        return ISyncSessionDriverChangeError;
      }
      [record setRecordIdentifier:[change recordIdentifier]];
      [[self recordLookup] setObject:record forKey:[change recordIdentifier]];
      break;
    case ISyncChangeTypeModify:
      record = [[self recordLookup] objectForKey:[change recordIdentifier]];
      break;
    default:
      errDict = [NSDictionary dictionaryWithObject:@"Unknown type"
                                            forKey:NSLocalizedDescriptionKey];
      *err = [NSError errorWithDomain:@"CIMGF" code:8001 userInfo:errDict];
      return ISyncSessionDriverChangeError;
  }
  for (NSDictionary *changeDict in [change changes]) {
    NSString *action = [changeDict valueForKey:ISyncChangePropertyActionKey];
    NSString *name = [changeDict valueForKey:ISyncChangePropertyNameKey];
    if ([name isEqualToString:kRecordEntityName]) {
      continue;
    }
    if ([action isEqualToString:ISyncChangePropertyClear]) {
      [record setNilValueForKey:name];
      continue;
    }
    id value = [changeDict valueForKey:ISyncChangePropertyValueKey];
    [record setValue:value forKey:name];
  }
  [record setUpdated:NO];
  return ISyncSessionDriverChangeAccepted;
}

-(NSDictionary*)recordsForEntityName:(NSString*)entityName 
													moreComing:(BOOL*)moreComing 
															 error:(NSError**)outError
{
  NSMutableDictionary *dict = [NSMutableDictionary dictionary];
  
  for (NSString *recordID in [[self recordLookup] allKeys]) {
    id object = [self objectForRecordIdentifier:recordID];
    NSString *objectEntityName = [[object class] entityName];
    if (![objectEntityName isEqualToString:entityName]) {
      continue;
    }
    [dict setValue:[object fullRecord] forKey:recordID];
  }
  
  *moreComing = NO;
  return dict;
}

-(NSURL*)clientDescriptionURL
{
	NSString *path = [[NSBundle mainBundle] pathForResource:@"ClientDescription" 
																									 ofType:@"plist"];
	return [NSURL fileURLWithPath:path];
}

//Less expensive for the sync engine but we have to know which properties
//changed to take advantage of it.

/*
-(NSArray*)changesForEntityName:(NSString*)entityName 
										 moreComing:(BOOL*)moreComing 
													error:(NSError**)outError
{
  NSLog(@"%@:%s entered", [self class], _cmd);
	return nil;
}*/

-(NSDictionary*)changedRecordsForEntityName:(NSString*)entityName 
																 moreComing:(BOOL*)moreComing 
																			error:(NSError**)outError
{
  NSMutableDictionary *dict = [NSMutableDictionary dictionary];
  for (NSString *key in [[self recordLookup] allKeys]) {
    id object = [[self recordLookup] objectForKey:key];
    if (![object updated]) continue;
    if (![[object entityName] isEqualToString:entityName]) continue;
    [dict setObject:[object fullRecord] forKey:key];
  }
	return dict;
}

#pragma mark -
#pragma mark OutlineView Delegate

- (BOOL)outlineView:(NSOutlineView*)outlineView 
shouldEditTableColumn:(NSTableColumn*)tableColumn 
							 item:(id)item
{
	id object = [item representedObject];
	NSLog(@"%@:%s object: %@", [self class], _cmd, object);
	if ([object isKindOfClass:[BookmarkEntity class]]) return YES;
	return NO;
}

- (NSArray*)tableSortDescriptors
{
  NSSortDescriptor *sorter = [[NSSortDescriptor alloc] initWithKey:@"position" 
                                                         ascending:YES];
  NSArray *array = [NSArray arrayWithObject:sorter];
  [sorter release];
  return array;
}

@end
